
// ***************************************************************************
//				    SBL2
//
// 
//
// 	TITLE: Schema Based Learning & Detour Behavior.
//  	GOAL:  Construction and Dynamics for SBL.
//

//	SECTIONS:
//	- Perceptual Schemas implemented as mappings.
//	- Schema Dynamics.


//	ARCHITECTURE: 
//	------------
//	Refer to Corbacho and Arbib (1995) for details.

//	DYNAMICS:
//	--------
//	Relaxation Labeling (Hummel and Zucker, 1983).


//	SIMPLIFICATIONS:
//	---------------
//	Perceptual Schemas as black boxes.
//	Motor Schemas as black boxes.


//	AUTHOR: Fernando Jose Corbacho
// 	DATE (last update): Jan 10th,96
//

// ***************************************************************************



#include "nsl_include.h"
#include "lib.h"   // Includes new Convolution and Integration.
#include "sbl_lib.h"

#include "math.h"
#include <stdio.h>

#include "2ddylink.h"	//4D kernels.
#include "dl2ps.h"	//display 4d masks.

//extern float rand();



NETWORK(SBL2);

//////////////////////////////// DIMENSIONS /////////////////////////////////


# define WX 200		// Dimensions of the world
# define WY 200

# define RECSIZE 80
# define TECTSIZE 80
# define TX 80		// # OF T cells.
# define TY 80			 // It SHOULD be 10: Convergence from 30 to 10.

# define RRF 19

# define F 9		// Tactile and Odor receptieve fields.

# define MAX pow(2,31)-1



nsl_vector maxim(nsl_vector& u)
{
	
	int imax = u.get_imax();
	nsl_vector c(imax);


	float max = 0;			// Maximum value
	int max_i=0;			// Coordinates of max value
	

/*
	for (int i=0; i < imax; i++)
		if (u.elem(i) > max) {
			c.elem(max_i) = 0;
			c.elem(i) = u.elem(i);
			max_i = i;
		}
		else 
			c.elem(i) = 0;

*/


	max = NSLmax(u);
	if (max == 0)
	 	for (int i=0; i < imax; i++)
			c.elem(i)= - 0.01;

	else
		for (int i=0; i < imax; i++)
			c.elem(i)= u.elem(i) - max + 0.01;


	c = NSLstep(c);
	return c;

}



nsl_matrix filter(nsl_matrix& view, int f)
{
	int i,j;
	int imax = view.get_imax();
	nsl_matrix c(imax, imax);

	for (i=0; i<imax; i++)
	  for (j=0; j<imax; j++)
		if (view.elem(i,j) == f)
			c.elem(i,j) = 1;
		else 
			c.elem(i,j) = 0;	

	return c;
}


//////////////////////// BOOTSTRAP ///////////////////////////

nsl_vector K(nsl_data A)
{
	nsl_vector c(13);

	int up = 6 + A.elem();		// center position + displacament.

	for (int i=0; i < 13; i++)	// initialize the kernel.
		c.elem(i) = 0;

	c.elem(up) = 1;			// where it is "active".



	return c;
}


nsl_vector T(nsl_vector& u, nsl_data A)
{

	int imax = u.get_imax();
	nsl_vector c(imax);
	nsl_vector kernel(13);


	kernel = K(A);
	c = kernel * u;

	return c;
}



float Match(nsl_vector& u, nsl_vector& u_d)
{

	int imax = u.get_imax();
	float c;
	c = 0;


	for (int i=0; i < imax; i++)
		c = c + abs (u.elem(i) - u_d.elem(i));

	return c;
}


nsl_matrix Posts_View()
{

	nsl_matrix c(TX, TY);

	c = 0;

	for (int j=0; j<69; j++)
	  for (int i=0; i<TY; i++)

		if ( (i > 2 && i < 8) || (i > 19 && i < 27) ||
		      (i > 51 && i < 59) || (i > 71 && i < 77) )

			c.elem(j,i) = 1;
			

	return c;	
}


nsl_matrix Post_View()
{

	nsl_matrix c(TX, TY);

	c = 0;

	for (int j=0; j<69; j++)
	  for (int i=0; i<TY; i++)

		if ( (i > 51 && i < 59) || (i > 71 && i < 77) )

			c.elem(j,i) = 1;
			

	return c;	
}


//////////////////////////////// ENVIRONMENT /////////////////////////////////


MATRIX(WorldXZ, WX, WY);   
MATRIX(WorldXY, WX, WY);

DATA(dd);	// increment in dots size for trajectory tracking.
DATA(gap_); float gap;

static float dt;
float boundary; DATA(boundary_);




//////////////////////////////// STIMULI ////////////////////////////////////

MATRIX(prey1, 5,10);		
MATRIX(predator1, 20,20);
MATRIX(rock1, 20,20);
MATRIX(hole1, 20,20);
MATRIX(fencepost, 25,1);		// several constitute a barrier.
MATRIX(post_up, 2,2);



DATA(yf); DATA(xf);			// COORDINATES of the frog.
					// -----------
DATA(yw); DATA(xw);			// coordinates of the prey.
DATA(ypr); DATA(xpr);			// coordinates of the predator.
DATA(xb); DATA(yb);			// coordinates of the barrier.
DATA(xb_init);				//
DATA(yh); DATA(xh);			// coordinates of the hole.



DATA(frog_speed);			// SPEED
					// -----
DATA(worm_speed);			//
DATA(step);				// step of the worm.
DATA(frog_forward);
DATA(move_trigger);

DATA(close);				// close to barrier
DATA(edge);				// the edge of barrier




// _________________________________________________________________________
// ___________________________ Brain DECLARATIONS __________________________



// ************************** SENSORY FIELDS **************************** 

MATRIX(Visual_field, RECSIZE, RECSIZE);

MATRIX(Tactile_field, F, F);

MATRIX(Odor_field, F, F);



// **************************** PREY RECOGNITION ************************ 
	
MATRIX(T5_2, TX, TY);
MATRIX(t5_2, TX, TY);			

MATRIX(t5_2_rf, 40, 40);		// KERNEL for Prey Attractant Field.
MATRIX(t5_2_erf, 40, 40); 
MATRIX(t5_2_irf, 40, 40);
DATA(t5_2_erf_dia); DATA(t5_2_erf_sig); DATA(t5_2_erf_wgt);




/* ************************* SOR RECOGNITION *************************** */
		
MATRIX(TH10, TX, TY);
MATRIX(th10, TX, TY);				

MATRIX(th10_erf, RRF, RRF);		// KERNEL for SO Repellent Field.
MATRIX(th10_irf, RRF, RRF);
MATRIX(th10_rf, RRF, RRF);

DATA(th10_erf_dia); DATA(th10_irf_dia);
DATA(th10_erf_sig); DATA(th10_irf_sig);
DATA(th10_erf_wgt); DATA(th10_irf_wgt);




/* *********************** "HOLE" RECOGNITION ************************* */





/* *********************** PREDATOR RECOGNITION ************************* */

		
MATRIX(TH3, TX, TY);
MATRIX(th3, TX, TY);				
MATRIX(th3_tm, 40, 40);	


/////////////////////// REACTIVE SCHEMAS ///////////////////

MATRIX(Prey_f, TX, TY);			// Prey Attractant Field
VECTOR(Prey_hor, TX);			// Parcellation

MATRIX(Pred_f, TX, TY);			// Predator Avoidance Field
VECTOR(Pred_hor, TX);			

VECTOR(TH10_hor, TX);			// Parcellation




/* *********************** SENSORYMOTOR Xform ******************** */
							

VECTOR(Gaps, TECTSIZE);			// openings within looming stimuli.
VECTOR(angle_position, TECTSIZE);
DATA(Angle);				// Orientation angle.
			  

					// KERNELS

four_d_matrix TH10_TM (TECTSIZE, 1, RRF, 1);

VECTOR(tm_erf, RRF); VECTOR(tm_irf, RRF); VECTOR(tm_rf, RRF);
DATA(tm_erf_dia); DATA(tm_irf_dia);
DATA(tm_erf_sig); DATA(tm_irf_sig);
DATA(tm_erf_wgt); DATA(tm_irf_wgt);
VECTOR(tm_rf_modul, RRF);	  	// modulated kernel.

VECTOR(Gradient, TECTSIZE);		// Gradient of weights.



VECTOR(u,TECTSIZE);			// for Winner Take All dynamics.
VECTOR(U,TECTSIZE);

VECTOR(IN, TECTSIZE);
VECTOR(modula1, TECTSIZE);
VECTOR(last_orientation,TECTSIZE);







/////////////////////////// SCHEMA    DIMENSIONS ////////////////////////////

# define Sch 	22			// # of SCHEMAS. Dynamically relocated.


# define L	5			// # of LABELS.

# define l_0	0 			// Inactive
# define l_1	1			// Gated
 
# define l_2	2 			// Different active signatures
# define l_3	3
# define l_4	4


/////////////////////////////////////////////////////////////////////////////
//////////////////////////////// SCHEMAS ////////////////////////////////////


VECTOR(Schemas, Sch);			// Schema Activity Variables.
DATA(th_sch);				// thresholds for schemas activation.
VECTOR(Support, Sch);			// Input to Schema.
DATA(noise);				// Noise on schema activations.


four_d_matrix R(Sch,Sch,L,L);		// COMPATIBILITY  Matrix.
MATRIX(R_show, Sch,Sch);		// display the 4D matrix.

MATRIX(P,Sch,L);
MATRIX(Q,Sch,L);


float sum;
int fo,pr,so,ta, dp;
int pa, pra, oa, ba;
int or,fw,sd,bk;

int sx;




/////////////////////////////////////////////////////////////////////////////
//////////////////////////// DUAL SPACES ////////////////////////////////////


VECTOR(Prey_d, TX);
VECTOR(Gaps_d, TX);
VECTOR(U_d, TX);


VECTOR(Incoherence, 6);


DATA(A);

// ____________________________________________________________________________
// _____________________________ MAIN PROGRAM _________________________________

INIT_MODULE(reset)
{

	dt = SBL2.get_delta();
	boundary = boundary_.elem();
	


////////////////////////////////////////////////////////////////////////////////

		///////// KERNELS ////////////




// PREY APPROACH FIELD
// ----
t5_2_erf = t5_2_erf_wgt * gauss2D(t5_2_erf_sig, t5_2_erf_dia, t5_2_erf_dia);
t5_2_rf = t5_2_erf;
					// Excitatory Gaussian.

// PREDATOR AVOIDANCE FIELD
// --------

	th3_tm = - t5_2_rf;		// Inhibitory Gaussian.


// OBSTACLE AVOIDANCE FIELD
// --------
	//INVERTED DOG MASKS FROM TH10 TO TEGMENTUM.

	// Create the default one.
	tm_erf = tm_erf_wgt * gauss1D(tm_erf_sig, tm_erf_dia.elem());
	tm_irf = tm_irf_wgt * gauss1D(tm_irf_sig, tm_irf_dia.elem());
	tm_rf = (tm_erf - tm_irf) ;

	tm_rf_modul = tm_rf;

	// Copy it around.
	for (int i=0; i <TECTSIZE; i++)
	  for (int j=0; j <RRF; j++)
	     	TH10_TM.elem(i,0).elem(j,0) = tm_rf.elem(j);




////////////////////////////////////////////////////////////////////////////////

		/////////  ENVIRONMENT ////////////

 

/* Barrier made out of fenceposts */
	
	gap = gap_.elem();
	xb = xb_init+30;

	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;


	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;


	int yp = yb.elem();
	int xp =  xb_init.elem();
	gap = gap + 2;

	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;


	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;



	WorldXY = WorldXY.put_sector(hole1,yh.elem(),xh.elem());
	WorldXZ = WorldXZ.put_sector(hole1,yh.elem(),xh.elem());
//		       		    -----







////////////////////////////////////////////////////////////////////////////////

		///////// SCHEMAS ////////////


	fo = 0; pr = 1; so = 2; ta = 3; dp = 4;


	pa = 8, pra = 9, oa = 10, ba = 11;
							sx = 15;

	or = 16; fw = 17; sd = 18; bk = 19;



	// Start with all inactive, i.e. LABEL 0


	for ( i=0; i < Sch; i++) 
		P.elem(i,0) = 0.9;

	
	// R.elem(i,j).elem(l1,l2)		COMPATIBILITY MATRIX INIT

	R.elem(fo,or).elem(l_0,l_0) = 1.0;	// food & Orient
	R.elem(fo,or).elem(l_0,l_1) = -1.0;
	R.elem(fo,or).elem(l_0,l_2) = -1.0;
	R.elem(fo,or).elem(l_0,l_3) = -1.0;
	R.elem(fo,or).elem(l_0,l_4) = -1.0;

	R.elem(fo,or).elem(l_2,l_0) = -1.0;	// food only gates motor schemas
	R.elem(fo,or).elem(l_2,l_1) = 1.0;
	
	R.elem(fo,or).elem(l_3,l_0) = -1.0;
	R.elem(fo,or).elem(l_3,l_1) = 1.0;

	R.elem(fo,or).elem(l_4,l_0) = -1.0;
	R.elem(fo,or).elem(l_4,l_1) = 1.0;



	R.elem(fo,fw).elem(l_0,l_0) = 1.0;	// food & Forward




	R.elem(so,fw).elem(l_0,l_0) = 1.0;	// SO & Forward




	R.elem(so,sd).elem(l_0,l_0) = 1.0;	// SO & Sidestep




	R.elem(pr,fw).elem(l_4,l_1) = 1.0;	// pr edator & forward




	R.elem(pr,sd).elem(l_4,l_1) = 1.0;	// pr & Sidestep



	R.elem(pr,bk).elem(l_4,l_1) = 1.0;	// pr & Back




	R.elem(ta,bk).elem(l_0,l_0) = 1.0;	// ta ctile & Back
	R.elem(ta,bk).elem(l_2,l_2) = 2.0;
	R.elem(ta,bk).elem(l_3,l_3) = 2.0;
	R.elem(ta,bk).elem(l_4,l_4) = 2.0;
	




/////////////////

	R.elem(fo,pa).elem(l_0,l_0) = 1.0;	// fo,pa
	R.elem(fo,pa).elem(l_2,l_2) = 1.0;
	R.elem(fo,pa).elem(l_3,l_3) = 1.0;
	R.elem(fo,pa).elem(l_4,l_4) = 1.0;

	R.elem(fo,pa).elem(l_0,l_2) = -1.0;
	R.elem(fo,pa).elem(l_0,l_3) = -1.0;
	R.elem(fo,pa).elem(l_0,l_4) = -1.0;
	
	R.elem(fo,pa).elem(l_2,l_3) = -1.0;
	R.elem(fo,pa).elem(l_2,l_4) = -1.0;

	R.elem(fo,pa).elem(l_3,l_2) = -1.0;
	R.elem(fo,pa).elem(l_3,l_4) = -1.0;

	R.elem(fo,pa).elem(l_4,l_2) = -1.0;
	R.elem(fo,pa).elem(l_4,l_3) = -1.0;




						// so,oa

	R.elem(so,oa).elem(l_0,l_0) = 1.0;
	R.elem(so,oa).elem(l_2,l_2) = 1.0;
	R.elem(so,oa).elem(l_3,l_3) = 1.0;
	R.elem(so,oa).elem(l_4,l_4) = 1.0;

	R.elem(so,oa).elem(l_0,l_2) = -1.0;
	R.elem(so,oa).elem(l_0,l_3) = -1.0;
	R.elem(so,oa).elem(l_0,l_4) = -1.0;
	
	R.elem(so,oa).elem(l_2,l_3) = -1.0;
	R.elem(so,oa).elem(l_2,l_4) = -1.0;

	R.elem(so,oa).elem(l_3,l_2) = -1.0;
	R.elem(so,oa).elem(l_3,l_4) = -1.0;

	R.elem(so,oa).elem(l_4,l_2) = -1.0;
	R.elem(so,oa).elem(l_4,l_3) = -1.0;



						// pr,pra

	R.elem(pr,pra).elem(l_0,l_0) = 1.0;
	R.elem(pr,pra).elem(l_2,l_2) = 1.0;
	R.elem(pr,pra).elem(l_3,l_3) = 1.0;
	R.elem(pr,pra).elem(l_4,l_4) = 1.0;

	R.elem(pr,pra).elem(l_0,l_2) = -1.0;
	R.elem(pr,pra).elem(l_0,l_3) = -1.0;
	R.elem(pr,pra).elem(l_0,l_4) = -1.0;
	
	R.elem(pr,pra).elem(l_2,l_3) = -1.0;
	R.elem(pr,pra).elem(l_2,l_4) = -1.0;

	R.elem(pr,pra).elem(l_3,l_2) = -1.0;
	R.elem(pr,pra).elem(l_3,l_4) = -1.0;

	R.elem(pr,pra).elem(l_4,l_2) = -1.0;
	R.elem(pr,pra).elem(l_4,l_3) = -1.0;




//////////////////
	R.elem(pa,fw).elem(l_2,l_2) = 1.0;

	R.elem(pa,or).elem(l_2,l_2) = 1.0;




	R.elem(oa,sd).elem(l_3,l_3) = 1.0;




	R.elem(pra,sd).elem(l_4,l_4) = 1.0;





	R.elem(ba,bk).elem(l_4,l_4) = 1.0;	// Bumping Avoidance
	R.elem(ba,sd).elem(l_4,l_4) = 1.0;




//////////////////




//////////////////
	R.elem(sx,or).elem(l_2,l_2) = 0.9;


//////////////////

	R.elem(fw,bk).elem(l_2,l_0) = 1;	// Reciprocal Inhibition.
	R.elem(fw,bk).elem(l_3,l_0) = 1;
	R.elem(fw,bk).elem(l_4,l_0) = 1;

	R.elem(bk,fw).elem(l_2,l_0) = 2;
	R.elem(bk,fw).elem(l_3,l_0) = 2;
	R.elem(bk,fw).elem(l_4,l_0) = 2;


}



RUN_MODULE(Update_World)
{


// ******* Place the  dynamic entities in the Environment   ******** 
			
	WorldXY = WorldXY.put_sector(prey1,yw.elem(), xw.elem());
	WorldXZ = WorldXZ.put_sector(prey1,yw.elem(), xw.elem());
//				     -----

	WorldXZ = WorldXZ.put_sector(predator1,ypr.elem(),xpr.elem());
	WorldXY = WorldXY.put_sector(predator1,ypr.elem(),xpr.elem());
//				     ---------


// *** Move prey in the Environment *** 

	if (xw.elem() < boundary ) 
		xw = xw + worm_speed.elem(); 	
	worm_speed = worm_speed * -1;		// Oscillatory behavior.




///////////////////    Update  Receptive Fields     ////////////////


	int xf_ = xf.elem();	//position of the frog
	int yf_ = yf.elem();
	

	Visual_field = WorldXZ.get_sector(yf_-RECSIZE+1, yf_,  xf_-RECSIZE/2+1, xf_+RECSIZE/2);

	Tactile_field = WorldXZ.get_sector(yf_-4,yf_+4,  xf_-4,xf_+4);





// *** Plot the trajectory of the frog. *** //
	// position of the frog.
	int x = xf.elem() ;
	int y = yf.elem() -25;

	WorldXY.elem(y,x) = 1 + dd.elem();


}



/* ***************************** SECTION 3 ********************************* */
/* *************************** BOOTSTRAP SCHEMAS  ************************** */

RUN_MODULE(Bootstrap)
{
// PREY-A-PREY

	Prey_d = T(Prey_hor, A);


// SOR-A-SOR

	Gaps_d = T(Gaps, A);

// PREDATOR-A-SOR


// TACTILE-A-TACTILE




// MHM-A-MHM

	U_d = T(U, A);




// PREY-MHM-TRIGEMINAL

// SOR-MHM-TACTILE



	A = 0;				// Reset motor feedback.

}




// ***************************************************************************
// ************************* SCHEMA MAPPINGS *********************************
// ***************************************************************************




// ************************ PERCEPTUAL SCHEMAS ****************************
// ************************************************************************



/////////////////////////// PREY RECOGNITION //////////////////////////



RUN_MODULE(net_T5_2)
// Element in the set?: if element in the set then T5_2 represents it.
{
	t5_2 = filter(Visual_field,2);
	if (t5_2.sum() > 1){
		P.elem(fo,l_2)=0.9;		// Label
		P.elem(fo,l_3)=0.1;			// Confidence
							// Temporal Structure
		}
	else {
		P.elem(fo, l_0)=1.0;
		
	     }
}

// PREY_RECOG
FOOD()
{
	T5_2 = NSLramp(t5_2);			// Output to other Schemas
	printf("Prey Recog	");

}



////////////////////////////// SOR /////////////////////////////////

RUN_MODULE(net_TH10)
{
        //th10 = filter(Visual_field, 1);

	if ( (close.elem() == 1) && (edge.elem() == 0) )
		th10 = Posts_View();
	else
		th10 = Post_View();
	


	if (th10.sum() > 1){
		P.elem(so,l_3)=0.9;		// Label
		P.elem(so,l_4)=0.1;
							// Confidence
							// Temporal Structure
		}
	else {
		P.elem(so, l_0)=1.0;
		
	     }
}


SOR()
{
	TH10 =	NSLramp(th10);  		// Output to other Schemas
	printf("SO Recog	");
}



////////////////////////////// PREDATOR /////////////////////////////////

RUN_MODULE(cell_TH3)
{
        th3 =	filter(Visual_field, 3);
	if (th3.sum() > 1)
	{
		P.elem(pr, l_4)=1;		// label
	}	
	else {
		P.elem(pr, l_0)=1.0;
		
	     }
}

// PREDATOR_RECOG
PREDATOR()
{
	TH3 = NSLramp(th3);
	printf("Predator Recog	");
}


////////////////////////////// TACTILE /////////////////////////////////

RUN_MODULE(Tact)
{

	if (Tactile_field.sum() > 1)
	{
		P.elem(ta, l_3)=1;			// label
	}
	else {
		P.elem(ta, l_0)=1.0;
		
	     }	
}


TACTILE()
{
	printf("Tactile		");
}


////////////////////////////// DEPTH /////////////////////////////////

RUN_MODULE(DEP)
{
/*
	if (yw - yf > 1)
	{
		P.elem(dp, l_3)=1;			// label
	}
*/	
}


DEPTH()
{
	printf("Depth		");
}


// ************************** REACTIVE SCHEMAS ****************************
// ************************************************************************



PREY_APPROACH()
{
	printf("Prey Approach	");
	Prey_f = newconv(t5_2_rf, T5_2);	
	Prey_hor = Prey_f.reduce_row();		// Parcellation	
}



SOR_AVOID()
{
	printf("SOR avoid	");
	TH10_hor = TH10.reduce_row();		// Parcellation
	Gaps = tm_rf_modul * TH10_hor;		//CONVOLVE 4D mask.
						//----------------
}


PREDATOR_A()
{
	printf("Pre avoid	");
	Pred_f = newconv(th3_tm, TH3);
	Pred_hor = Pred_f.reduce_row();		// Parcellation
	
}


BUMP_AVOID()
{
	printf("Bumping Avoidance	");

	
}








RUN_MODULE(Integrate)
{
	IN = IN + modula1;		//bumping reorientation.


//******		      	--       -----        --	
	DIFF.eq(u,1) = - u  +	 Gaps  +  Prey_hor  +  IN  ;
//******		      	--       -----        --	
	
	IN = 0;				//reset input.
}



RUN_MODULE(topol_to_bodyangle)
{	
	U = maxim(u);		
	angle_position = Gradient ^ U; 

//	-----
	Angle = int ( floor(angle_position.sum()) );	//eccentricity
//	-----
}




// **************************** MOTOR SCHEMAS *****************************
// ************************************************************************


RUN_MODULE(TM_to_Motor)
{
	
}


  FORWARD()
//---------
{

	printf("Forward		");

	// Condition for forward activation.
	

// Get the position from the motor map.


	if (yf.elem() > yw.elem())
		yf = yf - frog_forward.elem();


		WorldXY.elem(yf.elem(),xf.elem()) = 1;



	return(-1);
}



  SIDESTEP()
//----------
{

	printf("Sidestep	");

	// Translation keeping the angle.
	xf = xf + Angle;

A = Angle;

	return(-1);

}

//////////////////////////////// BACKUP ////////////////////////////////


RUN_MODULE(_Back_up)
{

	if (Tactile_field.sum() > 3)
		;		//remove
		//label;	// reflexive backup.
		//label   - FORWARD;

}


  BACKUP()
//----------
{

	printf("Backup	");

	yf.elem() = yf.elem()+frog_speed.elem();

	//Inhibit forward

	return(-1);
	
}



  ORIENT()
//--------
{

	printf("Orient		");
	return(-1);

}





/* ***************************** SECTION 4 ********************************* */
/* ***************************  COHERENCE MAX  ***************************** */

RUN_MODULE(Coherence)
{
// Pick out the maximum inconsistency: maximun measuring urgency and hamming 
// distance.


	// Incoherence.elem(0) = Match(Prey_hor,Prey_d);
	// Incoherence.elem(1) = Match(Gaps,Gaps_d);

	Incoherence.elem(5) = Match(U, U_d);


}




/* ***************************** SECTION 5 ********************************* */
/* *************************** SCHEMA DYNAMICS ***************************** */


RUN_MODULE(Triggering_Motor_schemas)
{
// This provides for activation, real position given by MHM.


	for (int i=0; i< Sch; i++)
		Schemas.elem(i) = P.elem(i,l_2) + P.elem(i,l_3) + P.elem(i,l_4);



// PERCEPTUAL

	if (Schemas.elem(fo) > th_sch.elem())			FOOD();
	if (Schemas.elem(pr) > th_sch.elem())			PREDATOR();
	if (Schemas.elem(so) > th_sch.elem())			SOR();
	if (Schemas.elem(ta) > th_sch.elem())			TACTILE();
	if (Schemas.elem(dp) > th_sch.elem())			DEPTH();


// REACTIVE

	if (Schemas.elem(pa) > th_sch.elem())			PREY_APPROACH();
	if (Schemas.elem(oa) > th_sch.elem())			SOR_AVOID();
	if (Schemas.elem(pra) > th_sch.elem())			PREDATOR_A();
	if (Schemas.elem(ba) > th_sch.elem())			BUMP_AVOID();

// MOTOR

	if (Schemas.elem(or) > th_sch.elem())			ORIENT();
	if (Schemas.elem(fw) > th_sch.elem())			FORWARD();
	if (Schemas.elem(sd) > th_sch.elem())			SIDESTEP();
	if (Schemas.elem(bk) > th_sch.elem())			BACKUP();

	
}





RUN_MODULE(Support_Vector)
{

	for (int i=0; i<Sch; i++){
	   for (int l1=0; l1<L; l1++){

		for (int j=0; j<Sch; j++){
		    for (int l2=0; l2<L; l2++){

		       	sum = sum + R.elem(j,i).elem(l2,l1) * P.elem(j,l2);

		    }
		}


	Q.elem(i,l1) = sum;
	sum = 0;

	    }
	}


}



RUN_MODULE(Update_Label)
{

	
	for (int i=0; i<Sch; i++){
	   for (int l1=0; l1<L; l1++){

		
		    for (int l2=0; l2<L; l2++){

		       	sum = sum + P.elem(i,l2) * (1 + Q.elem(i,l2));
		
		    }
		

				  // Introduce a little noise.
	P.elem(i,l1) = ( P.elem(i,l1) + noise.elem() * Q.elem(i,l1) ) 
					* (1+Q.elem(i,l1)) / sum;
	sum = 0;

	    }
	}


}







RUN_MODULE(Graphics)
{


	for (int i=0; i<Sch; i++)
	   for (int j=0; j<Sch;j++)

		R_show.elem(i,j) = R.elem(i,j).elem(l_1,l_2)-
R.elem(i,j).elem(l_2,l_1);



}











/*
RUN_MODULE(rank_incoherences)
{

	max = 5;
}

RUN_MODULE(construct_schema)
{
	if (Incoherence.elem(max) > 2)

	  for (i=0; i < TECTSIZE; i++){
		ASSOC.elem(a,C,i) = U_1.elem(i);	// Context C=0
		ASSOC.elem(a,P,i)= U.elem(i);		// Prediction P=1

		for (i=1;i<M;i++)			// Motor Pattern
			if (Schema.elem(i) > 1)
				A(a,i) = 1;	// connect to motor schemas.

		}

	}



}



RUN_MODULE(retrieving_past)
{

	// Parallel model matching.
	for (a=1; a< N; a++){
		Error(a) = Match(U, ASSOC(a,C));
		if (Error(a) < min)
			aa = a;
	}

	MODULA2 = ASSOC(aa,P);

}



*/
