
// ***************************************************************************
//			
//
//
// 	TITLE: Learning to Detour.
//  	GOAL:  Schema Construction and Dynamics towards SBL.
//
//	ARCHITECTURE: 
//	------------
//	Refer to Corbacho and Arbib (1995) for details.

//	Fields corresponding to perceptual schema instances;
//	Integration of fields; WTA; Sensorimotor Xform;
//	Motor schemas; Dynamics of schemas.




//	DYNAMICS:
//	--------
//	Relaxation Labeling (Hummel and Zucker, 1983).
//	Control of schemas based on distributed C&C dynamics.
//	Schemas by NN.


//	SIMPLIFICATIONS:
//	---------------
//	Perceptual Schemas (not Neural Implementation).
//	Motor Schemas.

//	NEXT:
//	----
//	Schema Construction in Run time.

//	THIS CODE IS ONLY PART OF THE WHOLE SIMULATION
//	e.g. "env.c"

//	AUTHOR: Fernando Jose Corbacho
// 	DATE (last update): Apr. 4th,97
//

// ***************************************************************************



#include "nsl_include.h"
#include "math.h"
#include <stdio.h>

#include "sbl_lib.h"		// Includes Gaussians (kernels) and different Convolutions.
#include "2ddylink.h"		// 4D kernels.
//#include "env.h"		// Environmental functions.






NETWORK(SBL2);

//////////////////////////////// DIMENSIONS /////////////////////////////////


# define WX 200		// Dimensions of the world
# define WY 200

# define RECSIZE 80
# define TECTSIZE 80
# define TX 80		// # OF T cells.
# define TY 80			 // It SHOULD be 10: Convergence from 30 to 10.

# define RRF 19		// tm_rf increased to 40.

# define F 9		// Tactile and Odor receptieve fields.

# define MAX pow(2,31)-1




/////////////////////////////// env.C //////////////////////////////////////

// Some environmental functions (e.g. for display).



nsl_matrix Central_View()
//        ---------------
{

	nsl_matrix c(TX, TY);

	c = 0;

	for (int j=0; j<69; j++)
	  for (int i=0; i<TY; i++)

		if ( (i > 2 && i < 8) || (i > 19 && i < 27) ||
		      (i > 51 && i < 59) || (i > 71 && i < 77) )

			c.elem(j,i) = 1;
			

	return c;	
}


nsl_matrix Edge_View()
//	   ------------
{

	nsl_matrix c(TX, TY);

	c = 0;

	for (int j=0; j<69; j++)
	  for (int i=0; i<TY; i++)

		if ( (i > 58 && i < 65) || (i > 73 && i < 77) ) 
		
						// || (i > 27 && i < 36)  )

			c.elem(j,i) = 1;
			

	return c;	
}



///////////////////////////////////////////////////////////////////////////




nsl_vector maxim(nsl_vector& u)
// returns the vector normalized (substraction) by its maximum.
// only the maximum is above threshold (by 0.01).
{
	
	int imax = u.get_imax();
	nsl_vector c(imax);


	float max = 0;			// Maximum value
	int max_i=0;			// Coordinates of max value
	


	max = NSLmax(u);
	if (max == 0)
	 	for (int i=0; i < imax; i++)
			c.elem(i)= - 0.01;

	else
		for (int i=0; i < imax; i++)
			c.elem(i)= u.elem(i) - max + 0.01;


	c = NSLstep(c);
	return c;

}



nsl_matrix filter(nsl_matrix& view, int f)
// filters out all the elements in the matrix that do not have a 
// value of "f". Simplified feature detector where "f" is the feature.

{
	int i,j;
	int imax = view.get_imax();
	nsl_matrix c(imax, imax);

	for (i=0; i<imax; i++)
	  for (j=0; j<imax; j++)
		if (view.elem(i,j) == f)
			c.elem(i,j) = 1;
		else 
			c.elem(i,j) = 0;	

	return c;
}



////////////////////////////// ENVIRONMENT /////////////////////////////////

MATRIX(WorldXZ, WX, WY);   
MATRIX(WorldXY, WX, WY);

DATA(dd);	// increment in dots size for trajectory tracking.


DATA(edge);				// the edge of barrier

static float dt;





DATA(gap_); float gap;
float boundary; DATA(boundary_);
DATA(safe_distance);


//////////////////////////////// STIMULI ////////////////////////////////////

MATRIX(prey1, 5,10);		
MATRIX(predator1, 20,20);
MATRIX(rock1, 20,20);
MATRIX(hole1, 20,20);

DATA(barrier);				// type of barrier.
MATRIX(fencepost, 25,1);		// several constitute a barrier.
MATRIX(post_up, 2,2);

					// COORDINATES
					// -----------
DATA(yw); DATA(xw);			// coordinates of the prey.
DATA(ypr); DATA(xpr);			// coordinates of the predator.
DATA(xb); DATA(yb);			// coordinates of the barrier.
DATA(xb_init);				//
DATA(yh); DATA(xh);			// coordinates of the hole.


DATA(worm_speed);			//
DATA(step);				// step of the worm.







//////////////////// FROG //////////////////

DATA(yf); DATA(xf);			// COORDINATES of the frog.
					// -----------
DATA(frog_speed);			// SPEED
					// -----
DATA(frog_forward);






// _________________________________________________________________________
// ___________________________ AAA DECLARATIONS __________________________



// ************************** RECEPTORS **************************** 

MATRIX(Visual_field, RECSIZE, RECSIZE);
MATRIX(Tactile_field, F, F);
DATA(Depth);


// **************************** PREY RECOGNITION ************************ 
	
MATRIX(T5_2, TX, TY);
MATRIX(t5_2, TX, TY);			


/* ************************* SOR RECOGNITION *************************** */
		
MATRIX(TH10, TX, TY);
MATRIX(th10, TX, TY);				




/////////////////////////// SENSORIMOTOR SCHEMAS ////////////////////////

MATRIX(Prey_f, TX, TY);			// Prey Attractant Field
VECTOR(Prey_hor, TX);			// Parcellation
DATA(prey_norm);			// Normalization
VECTOR(PREY_HOR, TX);

VECTOR(TH10_hor, TX);			// Parcellation
VECTOR(Gaps, TECTSIZE);			// openings within obstacles.
DATA(gaps_norm);			// Normalization
VECTOR(GAPS, TECTSIZE);

MATRIX(t5_2_rf, 40, 40);		// KERNEL for Prey Attractant Field.
MATRIX(t5_2_erf, 40, 40); 
MATRIX(t5_2_irf, 40, 40);
DATA(t5_2_erf_dia); DATA(t5_2_erf_sig); DATA(t5_2_erf_wgt);


					// KERNEL for Predator Avoid Field
VECTOR(tm_erf, RRF*2); VECTOR(tm_irf, RRF*2); VECTOR(tm_rf, RRF*2);
DATA(tm_erf_dia); DATA(tm_irf_dia);
DATA(tm_erf_sig); DATA(tm_irf_sig);
DATA(tm_erf_wgt); DATA(tm_irf_wgt);
VECTOR(tm_rf_modul, RRF*2);	  	// modulated kernel.



/* *********************** SENSORYMOTOR Xform ******************** */
							

VECTOR(angle_position, TECTSIZE);
DATA(Angle);				// Heading (target) angle.
			  
VECTOR(Gradient, TECTSIZE);		// Gradient of weights.



VECTOR(u,TECTSIZE);			// for Winner Take All dynamics.
VECTOR(U,TECTSIZE);

					// Other FIELDS.
					//	 =======
VECTOR(IN, TECTSIZE);
VECTOR(modula1, TECTSIZE);
VECTOR(Field_side, TECTSIZE);
DATA(Field_A);
DATA(Field_center);
VECTOR(last_orientation,TECTSIZE);


DATA(tune_tm);





//////////////////////////////// SCHEMAS ////////////////////////////////////


# define Sch 	22			// # of SCHEMAS. Dynamically relocated.

# define L	5			// # of LABELS.

# define l_0	0 			// Inactive
# define l_1	1			// Gated
 
# define l_2	2 			// Different active signatures
# define l_3	3
# define l_4	4



VECTOR(Schemas, Sch);			// Schema Activity Variables.
MATRIX(Schemas_disp, Sch, 1);
DATA(th_sch);				// thresholds for schemas activation.
VECTOR(Support, Sch);			// Input to Schema.
DATA(noise);				// Noise on schema activations.


four_d_matrix R(Sch,Sch,L,L);		// COMPATIBILITY  Matrix.
MATRIX(R_show, Sch,Sch);		// display the 4D matrix.

MATRIX(P,Sch,L);
MATRIX(Q,Sch,L);


float sum;
int fo,pr,so,ta, dp;			// fo stands for food, pr predator ...
int pa, pra, oa, ba;
int or,fw,sd,bk;

int sx;





// ____________________________________________________________________________
// _____________________________ MAIN PROGRAM _________________________________

INIT_MODULE(reset)
{

	dt = SBL2.get_delta();
	boundary = boundary_.elem();
	




		//////////////  KERNELS  //////////////


// PREY APPROACH KERNEL
// ----
	t5_2_erf = t5_2_erf_wgt * gauss2D(t5_2_erf_sig, t5_2_erf_dia, t5_2_erf_dia);
	t5_2_rf = t5_2_erf;
					// Excitatory Gaussian.



// OBSTACLE AVOIDANCE KERNEL
// --------
	
	
	tm_irf = tm_irf_wgt * gauss1D(tm_irf_sig, tm_irf_dia.elem());
	tm_rf = - tm_irf;

	tm_rf_modul = tm_rf;







		//////////////  SCHEMAS  ///////////////
							
	fo = 0; pr = 1; so = 2; ta = 3; dp = 4;			//INDICES
	pa = 8, pra = 9, oa = 10, ba = 11;
	sx = 15;
	or = 16; fw = 17; sd = 18; bk = 19;



	// Start with all inactive, i.e. LABEL 0

		for (int i=0; i < Sch; i++) 
			P.elem(i,0) = 0.9;

	

	
////////////////////// COMPATIBILITY MATRIX INIT  ///////////////////////	
	
	// Holds the Relations among different Schemas.
	
	
	
	R.elem(ta,ba).elem(l_2,l_2) = 3.0;	// tactile & Bumping Avoid
	R.elem(ta,ba).elem(l_3,l_3) = 3.0;
	R.elem(ta,ba).elem(l_4,l_4) = 3.0;
	
	R.elem(ta,ba).elem(l_0,l_0) = 3.0;	// reset quickly.




	R.elem(fo,pa).elem(l_0,l_0) = 1.0;	// fo,pa
	R.elem(fo,pa).elem(l_2,l_2) = 1.0;
	R.elem(fo,pa).elem(l_3,l_3) = 1.0;
	R.elem(fo,pa).elem(l_4,l_4) = 1.0;

	R.elem(fo,pa).elem(l_0,l_2) = -1.0;
	R.elem(fo,pa).elem(l_0,l_3) = -1.0;
	R.elem(fo,pa).elem(l_0,l_4) = -1.0;


						
	R.elem(so,oa).elem(l_0,l_0) = 1.0;	// so,oa
	R.elem(so,oa).elem(l_2,l_2) = 1.0;
	R.elem(so,oa).elem(l_3,l_3) = 1.0;
	R.elem(so,oa).elem(l_4,l_4) = 1.0;

	R.elem(so,oa).elem(l_0,l_2) = -1.0;
	R.elem(so,oa).elem(l_0,l_3) = -1.0;
	R.elem(so,oa).elem(l_0,l_4) = -1.0;
	



//////////////////

	R.elem(pa,fw).elem(l_2,l_2) = 1.0;	// support motor schemas

	R.elem(oa,sd).elem(l_3,l_3) = 1.0;
	
/*	
	R.elem(an,fw).elem(l_2,l_2) = 1.0;	// support motor schemas
	R.elem(an,sd).elem(l_3,l_3) = 1.0;
	R.elem(an,fw).elem(l_3,l_2) = - 1.0;	// support motor schemas
	R.elem(an,sd).elem(l_2,l_3) = - 1.0;
*/
	
	R.elem(ba,bk).elem(l_2,l_2) = 3.0;	// Bumping Avoid to Backup
	R.elem(ba,bk).elem(l_3,l_3) = 3.0;
	R.elem(ba,bk).elem(l_4,l_4) = 3.0;	// Fast "reflex".
	
	R.elem(ba,bk).elem(l_0,l_0) = 3.0;	


	R.elem(ba,sd).elem(l_2,l_2) = 0.5;	// Bumping Avoid to Sidestep
	R.elem(ba,sd).elem(l_3,l_3) = 0.5;
	R.elem(ba,sd).elem(l_4,l_4) = 0.5;
	

	


//////////////////

	R.elem(fw,bk).elem(l_2,l_0) = 1;	// Reciprocal Inhibition.
	R.elem(fw,bk).elem(l_3,l_0) = 1;
	R.elem(fw,bk).elem(l_4,l_0) = 1;

	R.elem(bk,fw).elem(l_2,l_0) = 2;
	R.elem(bk,fw).elem(l_3,l_0) = 2;
	R.elem(bk,fw).elem(l_4,l_0) = 2;







	/////////  ENVIRONMENT ////////////
////////////////////////////////////////////////////////////////////////////////

 

/* Barrier made out of fenceposts */
	
	gap = gap_.elem();
	xb = xb_init+30;

	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;

if (barrier.elem() == 20){

	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
	WorldXZ = WorldXZ.put_sector(fencepost,yb.elem(),xb.elem());
	xb = xb + gap;
}

	int yp = yb.elem() - 4;
	int xp =  xb_init.elem();
	gap = gap + 2;

	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;

if (barrier.elem() == 20){
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
	WorldXY = WorldXY.put_sector(post_up,yp,xp);
	xp = xp + gap;
}



////////////////////////////////////////////////////////////////////////////////
}




RUN_MODULE(Update_World)
{


// ******* Place the  dynamic entities in the Environment   ******** 
			
	WorldXY = WorldXY.put_sector(prey1,yw.elem(), xw.elem());
	WorldXZ = WorldXZ.put_sector(prey1,yw.elem(), xw.elem());
//				     -----

	WorldXZ = WorldXZ.put_sector(predator1,ypr.elem(),xpr.elem());
	//WorldXY = WorldXY.put_sector(predator1,ypr.elem(),xpr.elem());
//				     ---------


// *** Move prey in the Environment *** 

	if (xw.elem() < boundary ) 
		//xw = xw + worm_speed.elem(); 	
	worm_speed = worm_speed * -1;		// Oscillatory behavior.


//WorldXY = WorldXY.put_sector(hole1,yh.elem(),xh.elem());
//WorldXZ = WorldXZ.put_sector(hole1,yh.elem(),xh.elem());
//		       	       -----


///////////////////    Update  Receptive Fields     ////////////////


	int xf_ = xf.elem();	//position of the frog
	int yf_ = yf.elem();
	

	Visual_field = WorldXZ.get_sector(yf_-RECSIZE+1, yf_,  xf_-RECSIZE/2+1, xf_+RECSIZE/2);

//	Tactile_field = WorldXZ.get_sector(yf_-4,yf_+4,  xf_-4,xf_+4);



}



RUN_MODULE(Reset_next_iteration)
{
 	printf("\n");
 	printf("__________________________________________________________________\n");
 	
}




// ************************ PERCEPTUAL SCHEMAS ****************************
// ************************************************************************



/////////////////////////// PREY RECOGNITION //////////////////////////


RUN_MODULE(net_T5_2)
// Element in the set?: if element in the set then T5_2 represents it.
{
	t5_2 = filter(Visual_field,2);		// INSTANTIATED Schema.
	if (t5_2.sum() > 1){
		P.elem(fo,l_2)=0.9;		// Label
		P.elem(fo,l_3)=0.1;			// Confidence
							// Temporal Structure
		}
	else {
		P.elem(fo, l_0)=1.0;
		
	     }
}

// PREY_RECOG
FOOD()
{
	T5_2 = NSLramp(t5_2);			// Output to other Schemas
	printf("Prey Recog \t");		// ASSERTED Schema.

	return(1);
}



////////////////////////////// SOR /////////////////////////////////

RUN_MODULE(net_TH10)
{

// frog 20 cm from barrier.

  //if ( (yf.elem() - yb.elem()) > safe_distance.elem() * 4 )
  if ( (yf.elem() - yb.elem()) > 100 )		
        th10 = filter(Visual_field, 1);

	else if (  (xf.elem() > xb_init.elem() + edge.elem()) && 
		   (xf.elem() < xb.elem() - edge.elem())   )
			th10 = Central_View();
		
		else
			th10 = Edge_View();	// on the edge of the barrier.
	


	if (th10.sum() > 1){
		P.elem(so,l_3)=0.9;		// Label
		P.elem(so,l_4)=0.1;
							// Confidence
							// Temporal Structure
		}
	else {
		P.elem(so, l_0)=1.0;
		
	     }
	     
	     
   // Across the barrier
   	     
	      if ( yf.elem() < yb.elem()-8 ){
	      		P.elem(so, l_0)=1.0;
	      		P.elem(so, l_3)=0;
	      		th10 = 0;
	      		//xf = xf +10;
	      		}
	     
}


SOR()
{
	TH10 =	NSLramp(th10);  		// Output to other Schemas
	printf("SO Recog \t");
	
	return(1);
}



////////////////////////////// TACTILE /////////////////////////////////
RUN_MODULE(Tact)
{


	//if (Tactile_field.sum() > 1)

	if (yf.elem() < yb.elem() + safe_distance.elem())     
	{
		P.elem(ta, l_2)=4;			// Go up fast
		P.elem(ta, l_3)=4;
		P.elem(ta, l_4)=4;
		
	}
	else {
		P.elem(ta, l_0)=3.0;			// Go down fast.
			
		P.elem(ta, l_2)=0;			// Go up fast
		P.elem(ta, l_3)=0;
		P.elem(ta, l_4)=0;
		Schemas.elem(ta) = 0;  			 // Reset.
		
		
	     }	
	     
	     
// On the edge
   	     
	 if ( xf.elem() < xb_init.elem() ){
	      		
	      		printf("**************** WENT AROUND ! ************   \n");
	      		P.elem(ta, l_0)=3.0;			// Not Tactile
			P.elem(ta, l_2)=0;			// 
			P.elem(ta, l_3)=0;
			P.elem(ta, l_4)=0;	
			Schemas.elem(ta) = 0;  			 // Reset.
			printf("*******************************************\n");
	      		}
	      		
}



TACTILE()
{
	printf("Tactile \t");
	return(1);
}



// ************** *********** SENSORIMOTOR SCHEMAS *************************
// ************************************************************************



RUN_MODULE(Prey_Approach)
{

	Prey_f = t5_2_rf * T5_2;	//newconv(t5_2_rf, T5_2);	
//******Prey_hor = Prey_f.reduce_row();		// Parcellation: Only horizontal
						// 		component.
}

PREY_APPROACH()
{
	printf("Prey Appro \t");
	PREY_HOR = Prey_hor/NSLmax(Prey_hor);		// ASSERTED: Affect other schemas.		
	return(1);
}






RUN_MODULE(sor_avoidance)
{
	TH10_hor = TH10.reduce_row();		// Parcellation
	Gaps = tm_rf_modul * TH10_hor;		//CONVOLVE 4D mask.				
						//----------------
}

SOR_AVOID()
{
	printf("SOR avoid \t");
	GAPS = Gaps/NSLmax(Gaps*-1);			// ASSERTED: Affect other schemas.
	return(1);
}






BUMP_AVOID()
{
	printf("Bump Avoid \t");
	
	
	
// TUNING the KERNEL.

						// could use hebbian & dylinks.
	tm_irf_wgt = tm_irf_wgt + tune_tm;
	tm_irf_sig = tm_irf_sig + tune_tm;
	 
	tm_irf = tm_irf_wgt * gauss1D(tm_irf_sig, tm_irf_dia.elem());
	tm_rf = (tm_erf - tm_irf) ;

	tm_rf_modul = tm_rf;
	
	
	
// FIELD for SIDESTEP.
	
	
	Field_side.elem(Field_center.elem() ) = Field_A.elem();
	
	modula1 = Field_side;
	
	return(1);
	
}









RUN_MODULE(Integrate)
{
	IN = IN + modula1;		//bumping reorientation.


//******		      	--       -----        --	
//	DIFF.eq(u,1) = - u  +	 GAPS  +  PREY_HOR  +  IN  ;
//******		      	--       -----        --	
	 u  =  GAPS  +  PREY_HOR  +  IN  ;
	
	IN = 0;	modula1 = 0;			//reset input (cf. antidromic)
}



RUN_MODULE(topol_to_bodyangle)
{	
	U = maxim(u);		
	angle_position = Gradient ^ U; 

//	-----
	Angle = int ( floor(angle_position.sum()) );	//eccentricity
//	-----
}




// **************************** MOTOR SCHEMAS *****************************
// ************************************************************************


RUN_MODULE(TM_to_Motor)
{
	
}


  FORWARD()
//---------
{
	printf("Forward		");
	

// Get the position from the motor map.

	if (yf.elem() > yw.elem())
		yf = yf - frog_forward.elem();


		WorldXY.elem(yf.elem(),xf.elem()) = 1;		//Draw


	return(-1);
}



  SIDESTEP()
//----------
{

	printf("Sidestep: %4.1f     ",Angle.elem());

	
	xf = xf - Angle;

		WorldXY.elem(yf.elem(),xf.elem()) = 1;		//Draw
		
	return(-1);

}




  BACKUP()
//----------
{

	printf("Backup	");

	yf.elem() = yf.elem()+frog_speed.elem();

	WorldXY.elem(yf.elem(),xf.elem()) = 1;		//Draw

	return(-1);
	
}





/* ***************************** SECTION 5 ********************************* */
/* *************************** SCHEMA DYNAMICS ***************************** */




RUN_MODULE(Support_Vector)
// Defines the support that a Schema receives from all the other schemas.
// Similar to Hopfield dynamics (it is actually and extension comming from
// Relaxation Labeling (Hummel & Zucker, 1983).
// Notice that a Schema can be in "l" different states, i.e. schema labels.
// Hopfield corresponds to only two labels (i.e. ON, OFF).
{

	for (int i=0; i<Sch; i++){
	   for (int l1=0; l1<L; l1++){

		for (int j=0; j<Sch; j++){
		    for (int l2=0; l2<L; l2++){

		       	sum = sum + R.elem(j,i).elem(l2,l1) * P.elem(j,l2);

		    }
		}


	Q.elem(i,l1) = sum;
	sum = 0;

	    }
	}


}



RUN_MODULE(Update_Label)
// Updates the activity level of a schema based on the support it receices
// from the other schemas.
// Similar to Hopfield dynamics (it is actually and extension comming from
// Relaxation Labeling (Hummel & Zucker, 1983).
// Notice that a Schema can be in "l" different states, i.e. schema labels.
// Hopfield corresponds to only two labels (i.e. ON, OFF).
{

	
	for (int i=0; i<Sch; i++){
	   for (int l1=0; l1<L; l1++){

		
		    for (int l2=0; l2<L; l2++){

		       	sum = sum + P.elem(i,l2) * (1 + Q.elem(i,l2));
		
		    }
		

				  // Introduce a little noise.
	P.elem(i,l1) = ( P.elem(i,l1) + noise.elem() * Q.elem(i,l1) ) 
					* (1+Q.elem(i,l1)) / sum;
	sum = 0;

	    }
	}


}





RUN_MODULE(Triggering_schemas)
{
// This provides for activation, real position given by MHM.


	for (int i=0; i< Sch; i++)
		Schemas.elem(i) = P.elem(i,l_2) + P.elem(i,l_3) + P.elem(i,l_4);





	if ( Angle.elem() == 0 ){
		P.elem(fw,l_2) = P.elem(fw,l_2) + 0.3;	    	// ====
		Schemas.elem(sd) = Schemas.elem(sd) - 0.9;}
		
	else
		{Schemas.elem(fw) = Schemas.elem(fw) - 0.9;	// BIAS
							        // ====
		Schemas.elem(sd) = Schemas.elem(sd) + 0.3;}
 



// PERCEPTUAL

	if (Schemas.elem(fo) > th_sch.elem())			FOOD();
	//if (Schemas.elem(pr) > th_sch.elem())			PREDATOR();
	if (Schemas.elem(so) > th_sch.elem())			SOR();
	//if (Schemas.elem(dp) > th_sch.elem())			DEPTH();

	if (Schemas.elem(ta) > th_sch.elem())			
								{TACTILE();				
		//Schemas.elem(ta) = 0;  
		P.elem(ta,l_2) = 0;  P.elem(ta,l_3) = 0;		// Reset
	 	P.elem(ta,l_4) = 0;  P.elem(ta,l_0) = 3;}
	 	
	 	
	 	

// REACTIVE
printf("\n---------RS---------\n");
	if (Schemas.elem(pa) > th_sch.elem())			PREY_APPROACH();
	if (Schemas.elem(oa) > th_sch.elem())			SOR_AVOID();
	if (Schemas.elem(ba) > th_sch.elem())			BUMP_AVOID();



// MOTOR
// -----

printf("\n---------MS---------\n");
	
	
	if (Schemas.elem(fw) > th_sch.elem())			
								{FORWARD();				
		Schemas.elem(fw) = 0;  
		P.elem(fw,l_2) = 0;  P.elem(fw,l_3) = 0;		// Reset
	 	P.elem(fw,l_4) = 0;  P.elem(fw,l_0) = 2;}
	
	if (Schemas.elem(sd) > th_sch.elem())			{SIDESTEP();
	
		Schemas.elem(sd) = 0;  
		P.elem(sd,l_2) = 0;  P.elem(sd,l_3) = 0;		// Reset
	 	P.elem(sd,l_4) = 0;  P.elem(sd,l_0) = 2;}
	
	
	if (Schemas.elem(bk) > th_sch.elem())			{BACKUP();

		Schemas.elem(bk) = 0;  
		P.elem(bk,l_2) = 0;  P.elem(bk,l_3) = 0;		// Reset
	 	P.elem(bk,l_4) = 0;  P.elem(bk,l_0) = 2;}
	
}



RUN_MODULE(Display)
{

	for (int i=0; i< Sch; i++)
	
		Schemas_disp.elem(i,0) = Schemas.elem(i);
		
}		
		





